.changelog {
    cursor:text;
    /*margin-top:7px;*/
    font-family:Monaco, Consolas, 'DejaVu Sans Mono', 'Droid Sans Mono', Monospace;
    font-size:14px;
    padding:10px;
    border:1px solid #333333; /* #e0e8e5 */
    color:#333333;
    background-color:#eff3f2;
    word-wrap:break-word;
    max-width:640px;
    display: inline-block;
    text-align:left;
}

.wrap {
    white-space:pre-wrap;
}