body {
    padding-left:15px;
    padding:5px;
    background-color:White;
    color:Black;
}

h1 {
    font-size:1.7em;
    margin:0px;
    padding:0px;
}

hr {
    color:Black;
    background-color:Black;
    height:1px;
    border:0;
}

.tool {
    font-size:1.55em;
}

a.tool {
    text-decoration:none;
    color:Black;
}
